@@ -20,8 +20,8 @@ from .. import mlog
 from .. import compilers
 import json
 import subprocess
-from ..mesonlib import MesonException
-from ..mesonlib import get_compiler_for_source, classify_unity_sources
+from ..mesonlib import MesonException, OrderedSet
+from ..mesonlib import classify_unity_sources
 from ..mesonlib import File
 from ..compilers import CompilerArgs
 from collections import OrderedDict
@@ -37,11 +37,13 @@ class CleanTrees:
         self.trees = trees
 
 class InstallData:
-    def __init__(self, source_dir, build_dir, prefix, strip_bin, mesonintrospect):
+    def __init__(self, source_dir, build_dir, prefix, strip_bin,
+                 install_umask, mesonintrospect):
         self.source_dir = source_dir
         self.build_dir = build_dir
         self.prefix = prefix
         self.strip_bin = strip_bin
+        self.install_umask = install_umask
         self.targets = []
         self.headers = []
         self.man = []
@@ -52,6 +54,16 @@ class InstallData:
         self.install_subdirs = []
         self.mesonintrospect = mesonintrospect
 
+class TargetInstallData:
+    def __init__(self, fname, outdir, aliases, strip, install_name_mappings, install_rpath, install_mode):
+        self.fname = fname
+        self.outdir = outdir
+        self.aliases = aliases
+        self.strip = strip
+        self.install_name_mappings = install_name_mappings
+        self.install_rpath = install_rpath
+        self.install_mode = install_mode
+
 class ExecutableSerialisation:
     def __init__(self, name, fname, cmd_args, env, is_cross, exe_wrapper,
                  workdir, extra_paths, capture):
@@ -316,57 +328,61 @@ class Backend:
             self.write_benchmark_file(datafile)
         return test_data, benchmark_data
 
-    def determine_linker(self, target):
+    def determine_linker_and_stdlib_args(self, target):
         '''
         If we're building a static library, there is only one static linker.
         Otherwise, we query the target for the dynamic linker.
         '''
         if isinstance(target, build.StaticLibrary):
             if target.is_cross:
-                return self.build.static_cross_linker
+                return self.build.static_cross_linker, []
             else:
-                return self.build.static_linker
-        l = target.get_clike_dynamic_linker()
-        if not l:
-            m = "Couldn't determine linker for target {!r}"
-            raise MesonException(m.format(target.name))
-        return l
+                return self.build.static_linker, []
+        l, stdlib_args = target.get_clink_dynamic_linker_and_stdlibs()
+        return l, stdlib_args
 
-    def rpaths_for_bundled_shared_libraries(self, target):
+    @staticmethod
+    def _libdir_is_system(libdir, compilers):
+        for cc in compilers.values():
+            if libdir in cc.get_library_dirs():
+                return True
+        return False
+
+    def rpaths_for_bundled_shared_libraries(self, target, exclude_system=True):
         paths = []
         for dep in target.external_deps:
-            if isinstance(dep, (dependencies.ExternalLibrary, dependencies.PkgConfigDependency)):
-                la = dep.link_args
-                if len(la) == 1 and os.path.isabs(la[0]):
-                    # The only link argument is an absolute path to a library file.
-                    libpath = la[0]
-                    if libpath.startswith(('/usr/lib', '/lib')):
-                        # No point in adding system paths.
-                        continue
-                    if os.path.splitext(libpath)[1] not in ['.dll', '.lib', '.so']:
-                        continue
-                    absdir = os.path.dirname(libpath)
-                    if absdir.startswith(self.environment.get_source_dir()):
-                        rel_to_src = absdir[len(self.environment.get_source_dir()) + 1:]
-                        assert not os.path.isabs(rel_to_src), 'rel_to_src: {} is absolute'.format(rel_to_src)
-                        paths.append(os.path.join(self.build_to_src, rel_to_src))
-                    else:
-                        paths.append(absdir)
+            if not isinstance(dep, (dependencies.ExternalLibrary, dependencies.PkgConfigDependency)):
+                continue
+            la = dep.link_args
+            if len(la) != 1 or not os.path.isabs(la[0]):
+                continue
+            # The only link argument is an absolute path to a library file.
+            libpath = la[0]
+            libdir = os.path.dirname(libpath)
+            if exclude_system and self._libdir_is_system(libdir, target.compilers):
+                # No point in adding system paths.
+                continue
+            # Windows doesn't support rpaths, but we use this function to
+            # emulate rpaths by setting PATH, so also accept DLLs here
+            if os.path.splitext(libpath)[1] not in ['.dll', '.lib', '.so', '.dylib']:
+                continue
+            if libdir.startswith(self.environment.get_source_dir()):
+                rel_to_src = libdir[len(self.environment.get_source_dir()) + 1:]
+                assert not os.path.isabs(rel_to_src), 'rel_to_src: {} is absolute'.format(rel_to_src)
+                paths.append(os.path.join(self.build_to_src, rel_to_src))
+            else:
+                paths.append(libdir)
         return paths
 
     def determine_rpath_dirs(self, target):
         link_deps = target.get_all_link_deps()
-        result = []
+        result = set()
         for ld in link_deps:
             if ld is target:
                 continue
-            prospective = self.get_target_dir(ld)
-            if prospective not in result:
-                result.append(prospective)
-        for rp in self.rpaths_for_bundled_shared_libraries(target):
-            if rp not in result:
-                result += [rp]
-        return result
+            result.add(self.get_target_dir(ld))
+        result.update(self.rpaths_for_bundled_shared_libraries(target))
+        return list(result)
 
     def object_filename_from_source(self, target, source):
         assert isinstance(source, mesonlib.File)
@@ -511,7 +527,7 @@ class Backend:
         if not target.is_cross:
             # Compile args added from the env: CFLAGS/CXXFLAGS, etc. We want these
             # to override all the defaults, but not the per-target compile args.
-            commands += self.environment.coredata.external_args[compiler.get_language()]
+            commands += self.environment.coredata.get_external_args(compiler.get_language())
         # Always set -fPIC for shared libraries
         if isinstance(target, build.SharedLibrary):
             commands += compiler.get_pic_args()
@@ -568,28 +584,46 @@ class Backend:
             args.append(d_arg)
         return args
 
-    def determine_windows_extra_paths(self, target, extra_bdeps):
+    def get_mingw_extra_paths(self, target):
+        paths = OrderedSet()
+        # The cross bindir
+        root = self.environment.cross_info.get_root()
+        if root:
+            paths.add(os.path.join(root, 'bin'))
+        # The toolchain bindir
+        sys_root = self.environment.cross_info.get_sys_root()
+        if sys_root:
+            paths.add(os.path.join(sys_root, 'bin'))
+        # Get program and library dirs from all target compilers
+        if isinstance(target, build.BuildTarget):
+            for cc in target.compilers.values():
+                paths.update(cc.get_program_dirs())
+                paths.update(cc.get_library_dirs())
+        return list(paths)
+
+    def determine_windows_extra_paths(self, target, extra_bdeps, is_cross=False):
         '''On Windows there is no such thing as an rpath.
         We must determine all locations of DLLs that this exe
         links to and return them so they can be used in unit
         tests.'''
-        result = []
-        prospectives = []
-        if isinstance(target, build.Executable):
-            prospectives = target.get_transitive_link_deps()
+        result = set()
+        prospectives = set()
+        if isinstance(target, build.BuildTarget):
+            prospectives.update(target.get_transitive_link_deps())
             # External deps
-            for deppath in self.rpaths_for_bundled_shared_libraries(target):
-                result.append(os.path.normpath(os.path.join(self.environment.get_build_dir(), deppath)))
+            for deppath in self.rpaths_for_bundled_shared_libraries(target, exclude_system=False):
+                result.add(os.path.normpath(os.path.join(self.environment.get_build_dir(), deppath)))
         for bdep in extra_bdeps:
-            prospectives += bdep.get_transitive_link_deps()
+            prospectives.update(bdep.get_transitive_link_deps())
         # Internal deps
         for ld in prospectives:
             if ld == '' or ld == '.':
                 continue
             dirseg = os.path.join(self.environment.get_build_dir(), self.get_target_dir(ld))
-            if dirseg not in result:
-                result.append(dirseg)
-        return result
+            result.add(dirseg)
+        if is_cross:
+            result.update(self.get_mingw_extra_paths(target))
+        return list(result)
 
     def write_benchmark_file(self, datafile):
         self.write_test_serialisation(self.build.get_benchmarks(), datafile)
@@ -618,14 +652,20 @@ class Backend:
                 exe_wrapper = self.environment.cross_info.config['binaries'].get('exe_wrapper', None)
             else:
                 exe_wrapper = None
-            if mesonlib.is_windows() or mesonlib.is_cygwin():
-                extra_paths = self.determine_windows_extra_paths(exe, [])
+            if mesonlib.for_windows(is_cross, self.environment) or \
+               mesonlib.for_cygwin(is_cross, self.environment):
+                extra_bdeps = []
+                if isinstance(exe, build.CustomTarget):
+                    extra_bdeps = exe.get_transitive_build_target_deps()
+                extra_paths = self.determine_windows_extra_paths(exe, extra_bdeps, is_cross)
             else:
                 extra_paths = []
             cmd_args = []
             for a in t.cmd_args:
                 if hasattr(a, 'held_object'):
                     a = a.held_object
+                if isinstance(a, build.BuildTarget):
+                    extra_paths += self.determine_windows_extra_paths(a, [])
                 if isinstance(a, mesonlib.File):
                     a = os.path.join(self.environment.get_build_dir(), a.rel_to_builddir(self.build_to_src))
                     cmd_args.append(a)
@@ -710,7 +750,7 @@ class Backend:
         result = OrderedDict()
         # Get all build and custom targets that must be built by default
         for name, t in self.build.get_targets().items():
-            if t.build_by_default or t.install or t.build_always:
+            if t.build_by_default or t.install:
                 result[name] = t
         # Get all targets used as test executables and arguments. These must
         # also be built by default. XXX: Sometime in the future these should be
@@ -797,7 +837,7 @@ class Backend:
         outdir = self.get_target_dir(target)
         if absolute_outputs:
             source_root = self.environment.get_source_dir()
-            build_root = self.environment.get_source_dir()
+            build_root = self.environment.get_build_dir()
             outdir = os.path.join(self.environment.get_build_dir(), outdir)
         outputs = []
         for i in target.get_outputs():
