@@ -27,9 +27,9 @@ from pathlib import PurePath
 
 from .. import mlog
 from .. import mesonlib
-from ..mesonlib import (
-    MesonException, Popen_safe, version_compare_many, version_compare, listify
-)
+from ..compilers import clib_langs
+from ..mesonlib import MesonException, OrderedSet
+from ..mesonlib import Popen_safe, version_compare_many, version_compare, listify
 
 
 # These must be defined in this file to avoid cyclical references.
@@ -101,6 +101,9 @@ class Dependency:
         self.type_name = type_name
         self.compile_args = []
         self.link_args = []
+        # Raw -L and -l arguments without manual library searching
+        # If None, self.link_args will be used
+        self.raw_link_args = None
         self.sources = []
         self.methods = self._process_method_kw(kwargs)
 
@@ -111,7 +114,9 @@ class Dependency:
     def get_compile_args(self):
         return self.compile_args
 
-    def get_link_args(self):
+    def get_link_args(self, raw=False):
+        if raw and self.raw_link_args is not None:
+            return self.raw_link_args
         return self.link_args
 
     def found(self):
@@ -220,6 +225,7 @@ class ExternalDependency(Dependency):
             self.want_cross = not kwargs['native']
         else:
             self.want_cross = self.env.is_cross_build()
+        self.clib_compiler = None
         # Set the compiler that will be used by this dependency
         # This is only used for configuration checks
         if self.want_cross:
@@ -230,19 +236,20 @@ class ExternalDependency(Dependency):
         # else try to pick something that looks usable.
         if self.language:
             if self.language not in compilers:
-                m = self.name.capitalize() + ' requires a {} compiler'
+                m = self.name.capitalize() + ' requires a {0} compiler, but ' \
+                    '{0} is not in the list of project languages'
                 raise DependencyException(m.format(self.language.capitalize()))
-            self.compiler = compilers[self.language]
+            self.clib_compiler = compilers[self.language]
         else:
-            # Try to find a compiler that this dependency can use for compiler
-            # checks. It's ok if we don't find one.
-            for lang in ('c', 'cpp', 'objc', 'objcpp', 'fortran', 'd'):
-                self.compiler = compilers.get(lang, None)
-                if self.compiler:
+            # Try to find a compiler that can find C libraries for
+            # running compiler.find_library()
+            for lang in clib_langs:
+                self.clib_compiler = compilers.get(lang, None)
+                if self.clib_compiler:
                     break
 
     def get_compiler(self):
-        return self.compiler
+        return self.clib_compiler
 
     def get_partial_dependency(self, *, compile_args=False, link_args=False,
                                links=False, includes=False, sources=False):
@@ -436,8 +443,7 @@ class PkgConfigDependency(ExternalDependency):
                 if self.required:
                     raise DependencyException('Pkg-config binary missing from cross file')
             else:
-                pkgname = environment.cross_info.config['binaries']['pkgconfig']
-                potential_pkgbin = ExternalProgram(pkgname, silent=True)
+                potential_pkgbin = ExternalProgram.from_cross_info(environment.cross_info, 'pkgconfig')
                 if potential_pkgbin.found():
                     self.pkgbin = potential_pkgbin
                     PkgConfigDependency.class_pkgbin = self.pkgbin
@@ -518,8 +524,12 @@ class PkgConfigDependency(ExternalDependency):
                         self.version_reqs)
 
     def _call_pkgbin_real(self, args, env):
-        p, out = Popen_safe(self.pkgbin.get_command() + args, env=env)[0:2]
-        return p.returncode, out.strip()
+        cmd = self.pkgbin.get_command() + args
+        p, out = Popen_safe(cmd, env=env)[0:2]
+        rc, out = p.returncode, out.strip()
+        call = ' '.join(cmd)
+        mlog.debug("Called `{}` -> {}\n{}".format(call, rc, out))
+        return rc, out
 
     def _call_pkgbin(self, args, env=None):
         if env is None:
@@ -579,38 +589,76 @@ class PkgConfigDependency(ExternalDependency):
         libcmd = [self.name, '--libs']
         if self.static:
             libcmd.append('--static')
-            # Force pkg-config to output -L fields even if they are system
-            # paths so we can do manual searching with cc.find_library() later.
-            env = os.environ.copy()
-            env['PKG_CONFIG_ALLOW_SYSTEM_LIBS'] = '1'
+        # Force pkg-config to output -L fields even if they are system
+        # paths so we can do manual searching with cc.find_library() later.
+        env = os.environ.copy()
+        env['PKG_CONFIG_ALLOW_SYSTEM_LIBS'] = '1'
         ret, out = self._call_pkgbin(libcmd, env=env)
         if ret != 0:
             raise DependencyException('Could not generate libs for %s:\n\n%s' %
                                       (self.name, out))
-        self.link_args = []
-        libpaths = []
-        static_libs_notfound = []
+        # Also get the 'raw' output without -Lfoo system paths for usage when
+        # a library can't be found, and also in gnome.generate_gir
+        # + gnome.gtkdoc which need -L -l arguments.
+        ret, out_raw = self._call_pkgbin(libcmd)
+        if ret != 0:
+            raise DependencyException('Could not generate libs for %s:\n\n%s' %
+                                      (self.name, out_raw))
+        link_args = []
+        raw_link_args = []
+        # Library paths should be safe to de-dup
+        libpaths = OrderedSet()
+        raw_libpaths = OrderedSet()
+        # Track -lfoo libraries to avoid duplicate work
+        libs_found = OrderedSet()
+        # Track not-found libraries to know whether to add library paths
+        libs_notfound = []
+        libtype = 'static' if self.static else 'default'
+        # We always look for the file ourselves instead of depending on the
+        # compiler to find it with -lfoo or foo.lib (if possible) because:
+        # 1. We want to be able to select static or shared
+        # 2. We need the full path of the library to calculate RPATH values
+        #
+        # Libraries that are provided by the toolchain or are not found by
+        # find_library() will be added with -L -l pairs.
         for lib in self._convert_mingw_paths(shlex.split(out)):
-            # If we want to use only static libraries, we have to look for the
-            # file ourselves instead of depending on the compiler to find it
-            # with -lfoo or foo.lib. However, we can only do this if we already
-            # have some library paths gathered.
-            if self.static:
-                if lib.startswith('-L'):
-                    libpaths.append(lib[2:])
+            if lib.startswith(('-L-l', '-L-L')):
+                # These are D language arguments, add them as-is
+                pass
+            elif lib.startswith('-L'):
+                libpaths.add(lib[2:])
+                continue
+            elif lib.startswith('-l'):
+                # Don't resolve the same -lfoo argument again
+                if lib in libs_found:
                     continue
-                # FIXME: try to handle .la files in static mode too?
-                elif lib.startswith('-l'):
-                    args = self.compiler.find_library(lib[2:], self.env, libpaths, libtype='static')
-                    if not args or len(args) < 1:
-                        if lib in static_libs_notfound:
-                            continue
+                if self.clib_compiler:
+                    args = self.clib_compiler.find_library(lib[2:], self.env,
+                                                           list(libpaths), libtype)
+                # If the project only uses a non-clib language such as D, Rust,
+                # C#, Python, etc, all we can do is limp along by adding the
+                # arguments as-is and then adding the libpaths at the end.
+                else:
+                    args = None
+                if args:
+                    libs_found.add(lib)
+                    # Replace -l arg with full path to library if available
+                    # else, library is provided by the compiler and can't be resolved
+                    if not args[0].startswith('-l'):
+                        lib = args[0]
+                else:
+                    # Library wasn't found, maybe we're looking in the wrong
+                    # places or the library will be provided with LDFLAGS or
+                    # LIBRARY_PATH from the environment (on macOS), and many
+                    # other edge cases that we can't account for.
+                    #
+                    # Add all -L paths and use it as -lfoo
+                    if lib in libs_notfound:
+                        continue
+                    if self.static:
                         mlog.warning('Static library {!r} not found for dependency {!r}, may '
                                      'not be statically linked'.format(lib[2:], self.name))
-                        static_libs_notfound.append(lib)
-                    else:
-                        # Replace -l arg with full path to static library
-                        lib = args[0]
+                    libs_notfound.append(lib)
             elif lib.endswith(".la"):
                 shared_libname = self.extract_libtool_shlib(lib)
                 shared_lib = os.path.join(os.path.dirname(lib), shared_libname)
@@ -621,14 +669,24 @@ class PkgConfigDependency(ExternalDependency):
                     raise DependencyException('Got a libtools specific "%s" dependencies'
                                               'but we could not compute the actual shared'
                                               'library path' % lib)
-                lib = shared_lib
                 self.is_libtool = True
-            self.link_args.append(lib)
+                lib = shared_lib
+                if lib in link_args:
+                    continue
+            link_args.append(lib)
+        # Also store the raw link arguments, and store raw_libpaths
+        for lib in self._convert_mingw_paths(shlex.split(out_raw)):
+            if lib.startswith('-L') and not lib.startswith(('-L-l', '-L-L')):
+                raw_libpaths.add(lib[2:])
+            raw_link_args.append(lib)
+        # Set everything
+        self.link_args = link_args
+        self.raw_link_args = raw_link_args
         # Add all -Lbar args if we have -lfoo args in link_args
-        if static_libs_notfound:
+        if libs_notfound:
             # Order of -L flags doesn't matter with ld, but it might with other
             # linkers such as MSVC, so prepend them.
-            self.link_args = ['-L' + lp for lp in libpaths] + self.link_args
+            self.link_args = ['-L' + lp for lp in raw_libpaths] + self.link_args
 
     def get_pkgconfig_variable(self, variable_name, kwargs):
         options = ['--variable=' + variable_name, self.name]
@@ -742,6 +800,17 @@ class ExternalProgram:
             self.command = listify(command)
         else:
             self.command = self._search(name, search_dir)
+
+        # Set path to be the last item that is actually a file (in order to
+        # skip options in something like ['python', '-u', 'file.py']. If we
+        # can't find any components, default to the last component of the path.
+        self.path = self.command[-1]
+        for i in range(len(self.command) - 1, -1, -1):
+            arg = self.command[i]
+            if arg is not None and os.path.isfile(arg):
+                self.path = arg
+                break
+
         if not silent:
             if self.found():
                 mlog.log('Program', mlog.bold(name), 'found:', mlog.green('YES'),
@@ -753,6 +822,24 @@ class ExternalProgram:
         r = '<{} {!r} -> {!r}>'
         return r.format(self.__class__.__name__, self.name, self.command)
 
+    @staticmethod
+    def from_cross_info(cross_info, name):
+        if name not in cross_info.config['binaries']:
+            return NonExistingExternalProgram()
+        command = cross_info.config['binaries'][name]
+        if not isinstance(command, (list, str)):
+            raise MesonException('Invalid type {!r} for binary {!r} in cross file'
+                                 ''.format(command, name))
+        if isinstance(command, list):
+            if len(command) == 1:
+                command = command[0]
+        # We cannot do any searching if the command is a list, and we don't
+        # need to search if the path is an absolute path.
+        if isinstance(command, list) or os.path.isabs(command):
+            return ExternalProgram(name, command=command, silent=True)
+        # Search for the command using the specified string!
+        return ExternalProgram(command, silent=True)
+
     @staticmethod
     def _shebang_to_cmd(script):
         """
@@ -889,11 +976,7 @@ class ExternalProgram:
         return self.command[:]
 
     def get_path(self):
-        if self.found():
-            # Assume that the last element is the full path to the script or
-            # binary being run
-            return self.command[-1]
-        return None
+        return self.path
 
     def get_name(self):
         return self.name
@@ -925,7 +1008,7 @@ class ExternalLibrary(ExternalDependency):
             else:
                 mlog.log('Library', mlog.bold(name), 'found:', mlog.red('NO'))
 
-    def get_link_args(self, language=None):
+    def get_link_args(self, language=None, **kwargs):
         '''
         External libraries detected using a compiler must only be used with
         compatible code. For instance, Vala libraries (.vapi files) cannot be
@@ -938,7 +1021,7 @@ class ExternalLibrary(ExternalDependency):
         if (self.language == 'vala' and language != 'vala') or \
            (language == 'vala' and self.language != 'vala'):
             return []
-        return self.link_args
+        return super().get_link_args(**kwargs)
 
     def get_partial_dependency(self, *, compile_args=False, link_args=False,
                                links=False, includes=False, sources=False):
@@ -957,6 +1040,8 @@ class ExtraFrameworkDependency(ExternalDependency):
         self.required = required
         self.detect(name, path)
         if self.found():
+            self.compile_args = ['-I' + os.path.join(self.path, self.name, 'Headers')]
+            self.link_args = ['-F' + self.path, '-framework', self.name.split('.')[0]]
             mlog.log('Dependency', mlog.bold(name), 'found:', mlog.green('YES'),
                      os.path.join(self.path, self.name))
         else:
@@ -971,7 +1056,7 @@ class ExtraFrameworkDependency(ExternalDependency):
         for p in paths:
             for d in os.listdir(p):
                 fullpath = os.path.join(p, d)
-                if lname != d.split('.')[0].lower():
+                if lname != d.rsplit('.', 1)[0].lower():
                     continue
                 if not stat.S_ISDIR(os.stat(fullpath).st_mode):
                     continue
@@ -982,16 +1067,6 @@ class ExtraFrameworkDependency(ExternalDependency):
         if not self.found() and self.required:
             raise DependencyException('Framework dependency %s not found.' % (name, ))
 
-    def get_compile_args(self):
-        if self.found():
-            return ['-I' + os.path.join(self.path, self.name, 'Headers')]
-        return []
-
-    def get_link_args(self):
-        if self.found():
-            return ['-F' + self.path, '-framework', self.name.split('.')[0]]
-        return []
-
     def get_version(self):
         return 'unknown'
 
