@@ -33,6 +33,7 @@ from .base import (
 
 
 class MPIDependency(ExternalDependency):
+
     def __init__(self, environment, kwargs):
         language = kwargs.get('language', 'c')
         super().__init__('mpi', environment, language, kwargs)
@@ -254,7 +255,7 @@ class OpenMPDependency(ExternalDependency):
         super().__init__('openmp', environment, language, kwargs)
         self.is_found = False
         try:
-            openmp_date = self.compiler.get_define('_OPENMP', '', self.env, [], [self])
+            openmp_date = self.clib_compiler.get_define('_OPENMP', '', self.env, [], [self])
         except mesonlib.EnvironmentException as e:
             mlog.debug('OpenMP support not available in the compiler')
             mlog.debug(e)
@@ -262,7 +263,7 @@ class OpenMPDependency(ExternalDependency):
 
         if openmp_date:
             self.version = self.VERSIONS[openmp_date]
-            if self.compiler.has_header('omp.h', '', self.env, dependencies=[self]):
+            if self.clib_compiler.has_header('omp.h', '', self.env, dependencies=[self]):
                 self.is_found = True
             else:
                 mlog.log(mlog.yellow('WARNING:'), 'OpenMP found but omp.h missing.')
@@ -428,6 +429,7 @@ class Python3Dependency(ExternalDependency):
 
 
 class PcapDependency(ExternalDependency):
+
     def __init__(self, environment, kwargs):
         super().__init__('pcap', environment, None, kwargs)
 
@@ -464,8 +466,8 @@ class PcapDependency(ExternalDependency):
 
     @staticmethod
     def get_pcap_lib_version(ctdep):
-        return ctdep.compiler.get_return_value('pcap_lib_version', 'string',
-                                               '#include <pcap.h>', ctdep.env, [], [ctdep])
+        return ctdep.clib_compiler.get_return_value('pcap_lib_version', 'string',
+                                                    '#include <pcap.h>', ctdep.env, [], [ctdep])
 
 
 class CupsDependency(ExternalDependency):
