@@ -19,10 +19,10 @@ from pathlib import Path
 from .. import mesonlib
 from . import ExtensionModule
 from mesonbuild.modules import ModuleReturnValue
-from . import permittedSnippetKwargs
 from ..interpreterbase import (
     noPosargs, noKwargs, permittedKwargs,
-    InterpreterObject, InvalidArguments
+    InterpreterObject, InvalidArguments,
+    FeatureNew
 )
 from ..interpreter import ExternalProgramHolder
 from ..interpreterbase import flatten
@@ -49,6 +49,7 @@ def run_command(python, command):
 
 
 class PythonDependency(ExternalDependency):
+
     def __init__(self, python_holder, environment, kwargs):
         super().__init__('python', environment, None, kwargs)
         self.name = 'python'
@@ -122,7 +123,7 @@ class PythonDependency(ExternalDependency):
                 libname += self.variables['ABIFLAGS']
             libdirs = []
 
-        largs = self.compiler.find_library(libname, environment, libdirs)
+        largs = self.clib_compiler.find_library(libname, environment, libdirs)
 
         self.is_found = largs is not None
 
@@ -281,7 +282,7 @@ class PythonInstallation(ExternalProgramHolder, InterpreterObject):
         self.platform = run_command(python, "import sysconfig; print (sysconfig.get_platform())")
         self.is_pypy = json.loads(run_command(python, IS_PYPY_COMMAND))
 
-    @permittedSnippetKwargs(mod_kwargs)
+    @permittedKwargs(mod_kwargs)
     def extension_module(self, interpreter, state, args, kwargs):
         if 'subdir' in kwargs and 'install_dir' in kwargs:
             raise InvalidArguments('"subdir" and "install_dir" are mutually exclusive')
@@ -309,7 +310,7 @@ class PythonInstallation(ExternalProgramHolder, InterpreterObject):
         dep = PythonDependency(self, interpreter.environment, kwargs)
         return interpreter.holderify(dep)
 
-    @permittedSnippetKwargs(['pure', 'subdir'])
+    @permittedKwargs(['pure', 'subdir'])
     def install_sources(self, interpreter, state, args, kwargs):
         pure = kwargs.pop('pure', False)
         if not isinstance(pure, bool):
@@ -430,6 +431,8 @@ class PythonInstallation(ExternalProgramHolder, InterpreterObject):
 
 
 class PythonModule(ExtensionModule):
+
+    @FeatureNew('Python Module', '0.46.0')
     def __init__(self, *args, **kwargs):
         super().__init__(*args, **kwargs)
         self.snippets.add('find_installation')
@@ -447,7 +450,7 @@ class PythonModule(ExtensionModule):
         else:
             return None
 
-    @permittedSnippetKwargs(['required'])
+    @permittedKwargs(['required'])
     def find_installation(self, interpreter, state, args, kwargs):
         required = kwargs.get('required', True)
         if not isinstance(required, bool):
