@@ -18,22 +18,10 @@ import functools
 from . import mparser
 from . import coredata
 from . import mesonlib
-
+from . import compilers
 
 forbidden_option_names = coredata.get_builtin_options()
-forbidden_prefixes = {'c_',
-                      'cpp_',
-                      'd_',
-                      'rust_',
-                      'fortran_',
-                      'objc_',
-                      'objcpp_',
-                      'vala_',
-                      'csharp_',
-                      'swift_',
-                      'b_',
-                      'backend_',
-                      }
+forbidden_prefixes = [lang + '_' for lang in compilers.all_languages] + ['b_', 'backend_']
 
 def is_invalid_name(name):
     if name in forbidden_option_names:
@@ -63,16 +51,21 @@ def permitted_kwargs(permitted):
 
 optname_regex = re.compile('[^a-zA-Z0-9_-]')
 
-@permitted_kwargs({'value'})
+@permitted_kwargs({'value', 'yield'})
 def StringParser(name, description, kwargs):
-    return coredata.UserStringOption(name, description,
-                                     kwargs.get('value', ''), kwargs.get('choices', []))
+    return coredata.UserStringOption(name,
+                                     description,
+                                     kwargs.get('value', ''),
+                                     kwargs.get('choices', []),
+                                     kwargs.get('yield', coredata.default_yielding))
 
-@permitted_kwargs({'value'})
+@permitted_kwargs({'value', 'yield'})
 def BooleanParser(name, description, kwargs):
-    return coredata.UserBooleanOption(name, description, kwargs.get('value', True))
+    return coredata.UserBooleanOption(name, description,
+                                      kwargs.get('value', True),
+                                      kwargs.get('yield', coredata.default_yielding))
 
-@permitted_kwargs({'value', 'choices'})
+@permitted_kwargs({'value', 'yield', 'choices'})
 def ComboParser(name, description, kwargs):
     if 'choices' not in kwargs:
         raise OptionException('Combo option missing "choices" keyword.')
@@ -82,9 +75,28 @@ def ComboParser(name, description, kwargs):
     for i in choices:
         if not isinstance(i, str):
             raise OptionException('Combo choice elements must be strings.')
-    return coredata.UserComboOption(name, description, choices, kwargs.get('value', choices[0]))
+    return coredata.UserComboOption(name,
+                                    description,
+                                    choices,
+                                    kwargs.get('value', choices[0]),
+                                    kwargs.get('yield', coredata.default_yielding),)
 
-@permitted_kwargs({'value', 'choices'})
+
+@permitted_kwargs({'value', 'min', 'max', 'yield'})
+def IntegerParser(name, description, kwargs):
+    if 'value' not in kwargs:
+        raise OptionException('Integer option must contain value argument.')
+    return coredata.UserIntegerOption(name,
+                                      description,
+                                      kwargs.get('min', None),
+                                      kwargs.get('max', None),
+                                      kwargs['value'],
+                                      kwargs.get('yield', coredata.default_yielding))
+
+# FIXME: Cannot use FeatureNew while parsing options because we parse it before
+# reading options in project(). See func_project() in interpreter.py
+#@FeatureNew('array type option()', '0.44.0')
+@permitted_kwargs({'value', 'yield', 'choices'})
 def string_array_parser(name, description, kwargs):
     if 'choices' in kwargs:
         choices = kwargs['choices']
@@ -99,33 +111,31 @@ def string_array_parser(name, description, kwargs):
         value = kwargs.get('value', [])
     if not isinstance(value, list):
         raise OptionException('Array choices must be passed as an array.')
-    return coredata.UserArrayOption(name, description, value, choices=choices)
+    return coredata.UserArrayOption(name,
+                                    description,
+                                    value,
+                                    choices=choices,
+                                    yielding=kwargs.get('yield', coredata.default_yielding))
+
+@permitted_kwargs({'value', 'yield'})
+def FeatureParser(name, description, kwargs):
+    return coredata.UserFeatureOption(name,
+                                      description,
+                                      kwargs.get('value', 'enabled'),
+                                      yielding=kwargs.get('yield', coredata.default_yielding))
 
 option_types = {'string': StringParser,
                 'boolean': BooleanParser,
                 'combo': ComboParser,
+                'integer': IntegerParser,
                 'array': string_array_parser,
+                'feature': FeatureParser,
                 }
 
 class OptionInterpreter:
-    def __init__(self, subproject, command_line_options):
+    def __init__(self, subproject):
         self.options = {}
         self.subproject = subproject
-        self.sbprefix = subproject + ':'
-        self.cmd_line_options = {}
-        for o in command_line_options:
-            if self.subproject != '': # Strip the beginning.
-                # Ignore options that aren't for this subproject
-                if not o.startswith(self.sbprefix):
-                    continue
-            try:
-                (key, value) = o.split('=', 1)
-            except ValueError:
-                raise OptionException('Option {!r} must have a value separated by equals sign.'.format(o))
-            # Ignore subproject options if not fetching subproject options
-            if self.subproject == '' and ':' in key:
-                continue
-            self.cmd_line_options[key] = value
 
     def process(self, option_file):
         try:
@@ -178,6 +188,13 @@ class OptionInterpreter:
         if func_name != 'option':
             raise OptionException('Only calls to option() are allowed in option files.')
         (posargs, kwargs) = self.reduce_arguments(node.args)
+
+        # FIXME: Cannot use FeatureNew while parsing options because we parse
+        # it before reading options in project(). See func_project() in
+        # interpreter.py
+        #if 'yield' in kwargs:
+        #    FeatureNew('option yield', '0.45.0').use(self.subproject)
+
         if 'type' not in kwargs:
             raise OptionException('Option call missing mandatory "type" keyword argument')
         opt_type = kwargs.pop('type')
@@ -197,6 +214,4 @@ class OptionInterpreter:
         opt = option_types[opt_type](opt_name, kwargs.pop('description', ''), kwargs)
         if opt.description == '':
             opt.description = opt_name
-        if opt_name in self.cmd_line_options:
-            opt.set_value(self.cmd_line_options[opt_name])
         self.options[opt_name] = opt
